import logging
from huey.backends.dummy import DummyQueue, DummyDataStore
from huey.bin.config import BaseConfiguration


class Config(BaseConfiguration):
    QUEUE = DummyQueue('test-queue')
    RESULT_STORE = DummyDataStore('test-results')
    LOGFILE = None
    LOGLEVEL = logging.INFO
    PERIODIC = True
    MAX_DELAY = 60
    BACKOFF = 2



import datetime
import unittest

from huey.decorators import crontab


class SkewCrontabTestCase(unittest.TestCase):
    def test_crontab_month(self):
        # validates the following months, 1, 4, 7, 8, 9
        valids = [1, 4, 7, 8, 9]
        validate_m = crontab(month='1,4,*/6,8-9')
        
        for x in xrange(1, 13):
            res = validate_m(datetime.datetime(2011, x, 1))
            self.assertEqual(res, x in valids)
    
    def test_crontab_day(self):
        # validates the following days
        valids = [1, 4, 7, 8, 9, 13, 19, 25, 31]
        validate_d = crontab(day='*/6,1,4,8-9')
        
        for x in xrange(1, 32):
            res = validate_d(datetime.datetime(2011, 1, x))
            self.assertEqual(res, x in valids)
    
    def test_crontab_hour(self):
        # validates the following hours
        valids = [0, 1, 4, 6, 8, 9, 12, 18]
        validate_h = crontab(hour='8-9,*/6,1,4')
        
        for x in xrange(24):
            res = validate_h(datetime.datetime(2011, 1, 1, x))
            self.assertEqual(res, x in valids)
        
        edge = crontab(hour=0)
        self.assertTrue(edge(datetime.datetime(2011, 1, 1, 0, 0)))
        self.assertFalse(edge(datetime.datetime(2011, 1, 1, 12, 0)))
    
    def test_crontab_minute(self):
        # validates the following minutes
        valids = [0, 1, 4, 6, 8, 9, 12, 18, 24, 30, 36, 42, 48, 54]
        validate_m = crontab(minute='4,8-9,*/6,1')
        
        for x in xrange(60):
            res = validate_m(datetime.datetime(2011, 1, 1, 1, x))
            self.assertEqual(res, x in valids)
    
    def test_crontab_day_of_week(self):
        # validates the following days of week
        # jan, 1, 2011 is a saturday
        valids = [2, 4, 9, 11, 16, 18, 23, 25, 30]
        validate_dow = crontab(day_of_week='0,2')
        
        for x in xrange(1, 32):
            res = validate_dow(datetime.datetime(2011, 1, x))
            self.assertEqual(res, x in valids)
    
    def test_crontab_all_together(self):
        # jan 1, 2011 is a saturday
        # may 1, 2011 is a sunday
        validate = crontab(
            month='1,5',
            day='1,4,7',
            day_of_week='0,6',
            hour='*/4',
            minute='1-5,10-15,50'
        )
        
        self.assertTrue(validate(datetime.datetime(2011, 5, 1, 4, 11)))
        self.assertTrue(validate(datetime.datetime(2011, 5, 7, 20, 50)))
        self.assertTrue(validate(datetime.datetime(2011, 1, 1, 0, 1)))
        
        # fails validation on month
        self.assertFalse(validate(datetime.datetime(2011, 6, 4, 4, 11)))
        
        # fails validation on day
        self.assertFalse(validate(datetime.datetime(2011, 1, 6, 4, 11)))
        
        # fails validation on day_of_week
        self.assertFalse(validate(datetime.datetime(2011, 1, 4, 4, 11)))
        
        # fails validation on hour
        self.assertFalse(validate(datetime.datetime(2011, 1, 1, 1, 11)))
        
        # fails validation on minute
        self.assertFalse(validate(datetime.datetime(2011, 1, 1, 4, 6)))



class BaseQueue(object):
    """
    Base implementation for a Queue, all backends should subclass
    """
    
    # whether this backend blocks while waiting for new results or should be
    # polled by the consumer
    blocking = False
    
    def __init__(self, name, **connection):
        """
        Initialize the Queue - this happens once when the module is loaded
        
        :param name: A string representation of the name for this queue
        :param connection: Connection parameters for the queue
        """
        self.name = name
        self.connection = connection
    
    def write(self, data):
        """
        Push 'data' onto the queue
        """
        raise NotImplementedError
    
    def read(self):
        """
        Pop 'data' from the queue, returning None if no data is available --
        an empty queue should not raise an Exception!
        """
        raise NotImplementedError
    
    def flush(self):
        """
        Delete everything from the queue
        """
        raise NotImplementedError
    
    def __len__(self):
        """
        Used primarily in tests, but return the number of items in the queue
        """
        raise NotImplementedError


class BaseDataStore(object):
    """
    Base implementation for a data store
    """
    def __init__(self, name, **connection):
        """
        Initialize the data store - this happens once when the module is loaded
        """
        self.name = name
        self.connection = connection
    
    def put(self, key, value):
        raise NotImplementedError
    
    def get(self, key):
        raise NotImplementedError
    
    def flush(self):
        raise NotImplementedError



from huey.decorators import queue_command as _queue_command, periodic_command as _periodic_command, crontab
from huey.djhuey import invoker


def queue_command(fn):
    return _queue_command(invoker)(fn)

def periodic_command(crontab):
    return _periodic_command(invoker, crontab)



import datetime
import logging
import threading
import time
import unittest

from huey.backends.dummy import DummyQueue, DummyDataStore
from huey.decorators import queue_command
from huey.exceptions import QueueException
from huey.queue import Invoker, QueueCommand, PeriodicQueueCommand
from huey.registry import registry
from huey.utils import local_to_utc
from huey.bin.config import BaseConfiguration
from huey.bin.huey_consumer import load_config, Consumer, IterableQueue


# store some global state
state = {}

# create a queue, result store and invoker for testing
test_queue = DummyQueue('test-queue')
test_result_store = DummyDataStore('test-queue')
test_task_store = DummyDataStore('test-tasks')
test_invoker = Invoker(test_queue, test_result_store, test_task_store)

# create a dummy config for passing to the consumer
class DummyConfiguration(BaseConfiguration):
    QUEUE = test_queue
    RESULT_STORE = test_result_store
    TASK_STORE = test_result_store
    THREADS = 2

@queue_command(test_invoker)
def modify_state(k, v):
    state[k] = v
    return v

@queue_command(test_invoker)
def blow_up():
    raise Exception('blowed up')


# create a log handler that will track messages generated by the consumer
class TestLogHandler(logging.Handler):
    def __init__(self, *args, **kwargs):
        self.messages = []
        logging.Handler.__init__(self, *args, **kwargs)
        
    def emit(self, record):
        self.messages.append(record.getMessage())


class SkewConsumerTestCase(unittest.TestCase):
    def setUp(self):
        global state
        state = {}
        
        self.orig_sleep = time.sleep
        time.sleep = lambda x: None
        
        self.consumer = Consumer(test_invoker, DummyConfiguration)
        self.handler = TestLogHandler()
        self.consumer.logger.addHandler(self.handler)
    
    def tearDown(self):
        self.consumer.shutdown()
        self.consumer.logger.removeHandler(self.handler)
        time.sleep = self.orig_sleep
    
    def test_consumer_loader(self):
        config = load_config('huey.tests.config.Config')
        self.assertTrue(isinstance(config.QUEUE, DummyQueue))
        self.assertEqual(config.QUEUE.name, 'test-queue')
    
    def spawn(self, func, *args, **kwargs):
        t = threading.Thread(target=func, args=args, kwargs=kwargs)
        t.start()
        return t
    
    def test_iterable_queue(self):
        store = []
        q = IterableQueue()
        
        def do_queue(queue, result):
            for message in queue:
                result.append(message)
        
        t = self.spawn(do_queue, q, store)
        q.put(1)
        q.put(2)
        q.put(StopIteration)
        
        t.join()
        self.assertFalse(t.is_alive())
        self.assertEqual(store, [1, 2])
    
    def test_message_processing(self):
        self.consumer.start_message_receiver()
        self.consumer.start_worker_pool()
        
        self.assertFalse('k' in state)
        
        res = modify_state('k', 'v')
        res.get(blocking=True)
        
        self.assertTrue('k' in state)
        self.assertEqual(res.get(), 'v')
    
    def test_worker(self):
        res = modify_state('x', 'y')
        
        cmd = test_invoker.dequeue()
        self.assertEqual(res.get(), None)
        
        # we will be calling release() after finishing work
        self.consumer._pool.acquire()
        self.consumer.worker(cmd)
        
        self.assertTrue('x' in state)
        self.assertEqual(res.get(), 'y')
    
    def test_worker_exception(self):
        res = blow_up()
        cmd = test_invoker.dequeue()
        
        self.consumer._pool.acquire()
        self.consumer.worker(cmd)
        
        self.assertEqual(self.handler.messages, [
            'unhandled exception in worker thread',
        ])
    
    def test_pooling(self):
        # simulate acquiring two worker threads
        self.consumer._pool.acquire()
        self.consumer._pool.acquire()
        
        res = modify_state('x', 'y')
        
        # dequeue a *single* message 
        pt = self.spawn(self.consumer.check_message)
        
        # work on any messages generated by the processor thread
        st = self.spawn(self.consumer.worker_pool)
        
        # our result is not available since all workers are blocked
        self.assertEqual(res.get(), None)
        self.assertFalse(self.consumer._pool.acquire(blocking=False))
        
        # our processor is waiting
        self.assertTrue(pt.is_alive())
        self.assertEqual(self.consumer._queue.qsize(), 0)
        
        # release a worker
        self.consumer._pool.release()
        
        # we can get and block now, but will set a timeout of 3 to indicate that
        # something is wrong
        self.assertEqual(res.get(blocking=True, timeout=3), 'y')
        
        # this is done
        pt.join()
    
    def test_scheduling(self):
        dt = datetime.datetime(2011, 1, 1, 0, 0)
        dt2 = datetime.datetime(2037, 1, 1, 0, 0)
        r1 = modify_state.schedule(args=('k', 'v'), eta=dt, convert_utc=False)
        r2 = modify_state.schedule(args=('k2', 'v2'), eta=dt2, convert_utc=False)
        
        # dequeue a *single* message 
        pt = self.spawn(self.consumer.check_message)
        
        # work on any messages generated by the processor thread
        st = self.spawn(self.consumer.worker_pool)
        
        pt.join()
        self.assertTrue('k' in state)
        self.assertEqual(self.consumer.schedule._schedule, {})
        
        # dequeue a *single* message 
        pt = self.spawn(self.consumer.check_message)
        pt.join()
        
        # it got stored in the schedule instead of executing
        self.assertFalse('k2' in state)
        self.assertTrue(r2.task_id in self.consumer.schedule._schedule)
        
        # run through an iteration of the scheduler
        self.consumer.check_schedule(dt)
        
        # our command was not enqueued
        self.assertEqual(len(self.consumer.invoker.queue), 0)
        
        # try running the scheduler with the time the command should run
        self.consumer.check_schedule(dt2)
        
        # it was enqueued
        self.assertEqual(len(self.consumer.invoker.queue), 1)
        self.assertEqual(self.consumer.schedule._schedule, {})
        
        # dequeue and inspect -- it won't be executed because the scheduler will
        # see that it is scheduled to run in the future and plop it back into the
        # schedule
        command = self.consumer.invoker.dequeue()
        self.assertEqual(command.task_id, r2.task_id)
        self.assertEqual(command.execute_time, dt2)
    
    def test_schedule_local_utc(self):
        dt = datetime.datetime(2011, 1, 1, 0, 0)
        dt2 = datetime.datetime(2037, 1, 1, 0, 0)
        r1 = modify_state.schedule(args=('k', 'v'), eta=dt)
        r2 = modify_state.schedule(args=('k2', 'v2'), eta=dt2)
        
        # dequeue a *single* message 
        pt = self.spawn(self.consumer.check_message)
        
        # work on any messages generated by the processor thread
        st = self.spawn(self.consumer.worker_pool)
        
        pt.join()
        self.assertTrue('k' in state)
        self.assertEqual(self.consumer.schedule._schedule, {})
        
        # dequeue a *single* message 
        pt = self.spawn(self.consumer.check_message)
        pt.join()
        
        # it got stored in the schedule instead of executing
        self.assertFalse('k2' in state)
        self.assertTrue(r2.task_id in self.consumer.schedule._schedule)
        
        # run through an iteration of the scheduler
        self.consumer.check_schedule(dt)
        
        # our command was not enqueued
        self.assertEqual(len(self.consumer.invoker.queue), 0)
        
        # try running the scheduler with the time the command should run
        self.consumer.check_schedule(local_to_utc(dt2))
        
        # it was enqueued
        self.assertEqual(len(self.consumer.invoker.queue), 1)
        self.assertEqual(self.consumer.schedule._schedule, {})
        
        # dequeue and inspect -- it won't be executed because the scheduler will
        # see that it is scheduled to run in the future and plop it back into the
        # schedule
        command = self.consumer.invoker.dequeue()
        self.assertEqual(command.task_id, r2.task_id)
        self.assertEqual(command.execute_time, local_to_utc(dt2))
    
    def test_schedule_persistence(self):
        dt = datetime.datetime(2037, 1, 1, 0, 0)
        dt2 = datetime.datetime(2037, 1, 1, 0, 1)
        r = modify_state.schedule(args=('k', 'v'), eta=dt, convert_utc=False)
        r2 = modify_state.schedule(args=('k2', 'v2'), eta=dt2, convert_utc=False)
        
        # two messages in the queue
        self.assertEqual(len(self.consumer.invoker.queue), 2)
        
        # pull 'em down
        self.consumer.check_message()
        self.consumer.check_message()
        
        self.consumer.save_schedule()
        self.consumer.schedule._schedule = {}
        
        self.consumer.load_schedule()
        self.assertTrue(r.task_id in self.consumer.schedule._schedule)
        self.assertTrue(r2.task_id in self.consumer.schedule._schedule)
        
        cmd1 = self.consumer.schedule._schedule[r.task_id]
        cmd2 = self.consumer.schedule._schedule[r2.task_id]
        
        self.assertEqual(cmd1.execute_time, dt)
        self.assertEqual(cmd2.execute_time, dt2)
        
        # check w/conversion
        r3 = modify_state.schedule(args=('k3', 'v3'), eta=dt)
        self.consumer.check_message()
        
        self.consumer.save_schedule()
        self.consumer.schedule._schedule = {}
        
        self.consumer.load_schedule()
        cmd3 = self.consumer.schedule._schedule[r3.task_id]
        self.assertEqual(cmd3.execute_time, local_to_utc(dt))



import datetime
import os
import pickle
import uuid
import sys
import time

from huey.exceptions import QueueWriteException, QueueReadException, \
    DataStoreGetException, DataStorePutException, DataStoreTimeout
from huey.registry import registry
from huey.utils import wrap_exception, EmptyData


class AsyncData(object):
    def __init__(self, result_store, task_id):
        self.result_store = result_store
        self.task_id = task_id
        
        self._result = EmptyData
    
    def _get(self):
        if self._result is EmptyData:
            try:
                res = self.result_store.get(self.task_id)
            except:
                wrap_exception(DataStoreGetException)
            
            if res is not EmptyData:
                self._result = pickle.loads(res)
                return self._result
            else:
                return res
        else:
            return self._result
    
    def get(self, blocking=False, timeout=None, backoff=1.15, max_delay=1.0):
        if not blocking:
            res = self._get()
            if res is not EmptyData:
                return res
        else:
            start = time.time()
            delay = .1
            while self._result is EmptyData:
                if timeout and time.time() - start >= timeout:
                    raise DataStoreTimeout
                if delay > max_delay:
                    delay = max_delay
                if self._get() is EmptyData:
                    time.sleep(delay)
                    delay *= backoff
            
            return self._result


class Invoker(object):
    """
    The :class:`Invoker` is responsible for reading and writing to the queue
    and executing messages.  It talks to the :class:`CommandRegistry` to load
    up the proper :class:`QueueCommand` for each message
    """
    
    def __init__(self, queue, result_store=None, task_store=None, store_none=False,
                 always_eager=False):
        self.queue = queue
        self.result_store = result_store
        self.task_store = task_store
        self.blocking = self.queue.blocking
        self.store_none = store_none
        self.always_eager = always_eager
    
    def write(self, msg):
        try:
            self.queue.write(msg)
        except:
            wrap_exception(QueueWriteException)
    
    def enqueue(self, command):
        if self.always_eager:
            return command.execute()
        
        self.write(registry.get_message_for_command(command))
        
        if self.result_store:
            return AsyncData(self.result_store, command.task_id)
    
    def read(self):
        try:
            return self.queue.read()
        except:
            wrap_exception(QueueReadException)
    
    def dequeue(self):
        message = self.read()
        if message:
            return registry.get_command_for_message(message)
    
    def execute(self, command):
        if not isinstance(command, QueueCommand):
            raise TypeError('Unknown object: %s' % command)
        
        result = command.execute()

        if result is None and not self.store_none:
            return
        
        if self.result_store and not isinstance(command, PeriodicQueueCommand):
            serialized = pickle.dumps(result)
            try:
                self.result_store.put(command.task_id, serialized)
            except:
                wrap_exception(DataStorePutException)
        
        return result
    
    def flush(self):
        self.queue.flush()


class CommandSchedule(object):
    def __init__(self, invoker, key_name='schedule'):
        self.invoker = invoker
        self.key_name = key_name
        
        self.task_store = self.invoker.task_store
        self._schedule = {}
    
    def load(self):
        if self.task_store:
            serialized = self.task_store.get(self.key_name)
            
            if serialized:
                self.load_commands(pickle.loads(serialized))
    
    def load_commands(self, messages):
        for cmd_string in messages:
            try:
                cmd_obj = registry.get_command_for_message(cmd_string)
                self.add(cmd_obj)
            except QueueException:
                pass
    
    def save(self):
        if self.task_store:
            self.task_store.put(self.key_name, self.serialize_commands())
    
    def serialize_commands(self):
        messages = [registry.get_message_for_command(c) for c in self.commands()]
        return pickle.dumps(messages)
    
    def commands(self):
        return self._schedule.values()
    
    def should_run(self, cmd, dt=None):
        if cmd.execute_time is None:
            return True
        else:
            return cmd.execute_time <= (dt or datetime.datetime.now())
    
    def add(self, cmd):
        if not self.is_pending(cmd):
            self._schedule[cmd.task_id] = cmd
    
    def remove(self, cmd):
        if self.is_pending(cmd):
            del(self._schedule[cmd.task_id])
    
    def is_pending(self, cmd):
        return cmd.task_id in self._schedule


class QueueCommandMetaClass(type):
    def __init__(cls, name, bases, attrs):
        """
        Metaclass to ensure that all command classes are registered
        """
        registry.register(cls)


class QueueCommand(object):
    """
    A class that encapsulates the logic necessary to 'do something' given some
    arbitrary data.  When enqueued with the :class:`Invoker`, it will be
    stored in a queue for out-of-band execution via the consumer.  See also
    the :func:`queue_command` decorator, which can be used to automatically
    execute any function out-of-band.
    
    Example::
    
    class SendEmailCommand(QueueCommand):
        def execute(self):
            data = self.get_data()
            send_email(data['recipient'], data['subject'], data['body'])
    
    invoker.enqueue(
        SendEmailCommand({
            'recipient': 'somebody@spam.com',
            'subject': 'look at this awesome website',
            'body': 'http://youtube.com'
        })
    )
    """
    
    __metaclass__ = QueueCommandMetaClass
    
    def __init__(self, data=None, task_id=None, execute_time=None):
        """
        Initialize the command object with a receiver and optional data.  The
        receiver object *must* be a django model instance.
        """
        self.set_data(data)
        self.task_id = task_id or self.create_id()
        self.execute_time = execute_time
    
    def create_id(self):
        return str(uuid.uuid4())

    def get_data(self):
        """Called by the Invoker when a command is being enqueued"""
        return self.data

    def set_data(self, data):
        """Called by the Invoker when a command is dequeued"""
        self.data = data

    def execute(self):
        """Execute any arbitary code here"""
        raise NotImplementedError
    
    def __eq__(self, rhs):
        return \
            self.task_id == rhs.task_id and \
            self.execute_time == rhs.execute_time and \
            type(self) == type(rhs)


class PeriodicQueueCommand(QueueCommand):
    def create_id(self):
        return registry.command_to_string(type(self))
    
    def validate_datetime(self, dt):
        """Validate that the command should execute at the given datetime"""
        return False



import datetime
import unittest

from huey.backends.dummy import DummyQueue, DummyDataStore
from huey.decorators import queue_command, periodic_command, crontab
from huey.exceptions import QueueException
from huey.queue import Invoker, QueueCommand, PeriodicQueueCommand, CommandSchedule
from huey.registry import registry
from huey.utils import EmptyData, local_to_utc


queue_name = 'test-queue'
queue = DummyQueue(queue_name)
invoker = Invoker(queue)

res_queue_name = 'test-queue-2'
res_queue = DummyQueue(res_queue_name)
res_store = DummyDataStore(res_queue_name)
task_store = DummyDataStore(res_queue_name)

res_invoker = Invoker(res_queue, res_store, task_store)
res_invoker_nones = Invoker(res_queue, res_store, task_store, store_none=True)

# store some global state
state = {}

# create a decorated queue command
@queue_command(invoker)
def add(key, value):
    state[key] = value

# create a periodic queue command
@periodic_command(invoker, crontab(minute='0'))
def add_on_the_hour():
    state['periodic'] = 'x'

# define a command using the class
class AddCommand(QueueCommand):
    def execute(self):
        k, v = self.data
        state[k] = v

# create a command that raises an exception
class BampfException(Exception):
    pass

@queue_command(invoker)
def throw_error():
    raise BampfException('bampf')

@queue_command(res_invoker)
def add2(a, b):
    return a + b

@queue_command(res_invoker)
def returns_none():
    return None

@queue_command(res_invoker_nones)
def returns_none2():
    return None

class SkewTestCase(unittest.TestCase):
    def setUp(self):
        global state
        queue.flush()
        state = {}
    
    def test_registration(self):
        self.assertTrue('huey.tests.queue.queuecmd_add' in registry)
        self.assertTrue('huey.tests.queue.queuecmd_add_on_the_hour' in registry)
        self.assertTrue('huey.tests.queue.AddCommand' in registry)
    
    def test_enqueue(self):
        # sanity check
        self.assertEqual(len(queue), 0)
        
        # initializing the command does not enqueue it
        ac = AddCommand(('k', 'v'))
        self.assertEqual(len(queue), 0)
        
        # ok, enqueue it, then check that it was enqueued
        invoker.enqueue(ac)
        self.assertEqual(len(queue), 1)
        
        # it can be enqueued multiple times
        invoker.enqueue(ac)
        self.assertEqual(len(queue), 2)
        
        # no changes to state
        self.assertFalse('k' in state)
    
    def test_enqueue_decorator(self):
        # sanity check
        self.assertEqual(len(queue), 0)
        
        add('k', 'v')
        self.assertEqual(len(queue), 1)
        
        add('k', 'v')
        self.assertEqual(len(queue), 2)
        
        # no changes to state
        self.assertFalse('k' in state)
    
    def test_schedule(self):
        dt = datetime.datetime(2011, 1, 1, 0, 1)
        add('k', 'v')
        self.assertEqual(len(queue), 1)
        
        cmd = invoker.dequeue()
        self.assertEqual(cmd.execute_time, None)
        
        add.schedule(args=('k2', 'v2'), eta=dt)
        self.assertEqual(len(queue), 1)
        cmd = invoker.dequeue()
        self.assertEqual(cmd.execute_time, local_to_utc(dt))
        
        add.schedule(args=('k3', 'v3'), eta=dt, convert_utc=False)
        self.assertEqual(len(queue), 1)
        cmd = invoker.dequeue()
        self.assertEqual(cmd.execute_time, dt)
    
    def test_error_raised(self):
        throw_error()
        
        # no error
        cmd = invoker.dequeue()
        
        # error
        self.assertRaises(BampfException, invoker.execute, cmd)
    
    def test_dequeueing(self):
        res = invoker.dequeue() # no error raised if queue is empty
        self.assertEqual(res, None)
                
        add('k', 'v')
        cmd = invoker.dequeue()
        
        self.assertTrue(isinstance(cmd, QueueCommand))
        self.assertEqual(cmd.get_data(), (('k', 'v'), {}))
    
    def test_execution(self):
        self.assertFalse('k' in state)
        add('k', 'v')
        
        cmd = invoker.dequeue()
        self.assertFalse('k' in state)
        
        invoker.execute(cmd)
        self.assertEqual(state['k'], 'v')
        
        add('k', 'X')
        self.assertEqual(state['k'], 'v')
        
        invoker.execute(invoker.dequeue())
        self.assertEqual(state['k'], 'X')
        
        self.assertRaises(TypeError, invoker.execute, invoker.dequeue())
    
    def test_result_store(self):
        res = add2(1, 2)
        res2 = add2(4, 5)
        res3 = add2(0, 0)
        
        # none have been executed as yet
        self.assertEqual(res.get(), None)
        self.assertEqual(res2.get(), None)
        self.assertEqual(res3.get(), None)

        # execute the first task
        res_invoker.execute(res_invoker.dequeue())
        self.assertEqual(res.get(), 3)
        self.assertEqual(res2.get(), None)
        self.assertEqual(res3.get(), None)
        
        # execute the second task
        res_invoker.execute(res_invoker.dequeue())
        self.assertEqual(res.get(), 3)
        self.assertEqual(res2.get(), 9)
        self.assertEqual(res3.get(), None)
        
        # execute the 3rd, which returns a zero value
        res_invoker.execute(res_invoker.dequeue())
        self.assertEqual(res.get(), 3)
        self.assertEqual(res2.get(), 9)
        self.assertEqual(res3.get(), 0)
    
        # check that it returns None when nothing is present
        res = returns_none()
        self.assertEqual(res.get(), None)
        
        # execute, it will still return None, but underneath it is an EmptyResult
        # indicating its actual result was not persisted
        res_invoker.execute(res_invoker.dequeue())
        self.assertEqual(res.get(), None)
        self.assertEqual(res._result, EmptyData)

        # execute again, this time note that we're pointing at the invoker
        # that *does* accept None as a store-able result
        res = returns_none2()
        self.assertEqual(res.get(), None)

        # it stores None
        res_invoker_nones.execute(res_invoker_nones.dequeue())
        self.assertEqual(res.get(), None)
        self.assertEqual(res._result, None)
    
    def test_task_store(self):
        schedule = CommandSchedule(res_invoker)
        
        dt1 = datetime.datetime(2011, 1, 1, 0, 0)
        dt2 = datetime.datetime(2035, 1, 1, 0, 0)
        
        add2.schedule(args=('k', 'v'), eta=dt1)
        cmd1 = res_invoker.dequeue()
        
        add2.schedule(args=('k2', 'v2'), eta=dt2)
        cmd2 = res_invoker.dequeue()
        
        add2('k3', 'v3')
        cmd3 = res_invoker.dequeue()
        
        # add the command to the schedule
        schedule.add(cmd1)
        self.assertEqual(schedule.commands(), [cmd1])
        
        # cmd1 is in the schedule, cmd2 is not
        self.assertTrue(schedule.is_pending(cmd1))
        self.assertFalse(schedule.is_pending(cmd2))
        
        # multiple adds do not result in the list growing
        schedule.add(cmd1)
        self.assertEqual(schedule.commands(), [cmd1])
        
        # removing a non-existant cmd does not error
        schedule.remove(cmd2)
        
        # add a future-dated command
        schedule.add(cmd2)
        
        # sanity check what should be run
        self.assertTrue(schedule.should_run(cmd1))
        self.assertFalse(schedule.should_run(cmd2))
        self.assertTrue(schedule.should_run(cmd3))
        
        # check remove works
        schedule.remove(cmd1)
        self.assertEqual(schedule.commands(), [cmd2])
        
        # check saving works
        schedule.save()
        schedule._schedule = {}
        self.assertEqual(schedule.commands(), [])
        
        # check loading works
        schedule.load()
        self.assertEqual(schedule.commands(), [cmd2])
        
        schedule.add(cmd1)
        schedule.add(cmd3)
        schedule.save()
        schedule._schedule = {}
        
        schedule.load()
        self.assertEqual(len(schedule.commands()), 3)
        self.assertTrue(cmd1 in schedule.commands())
        self.assertTrue(cmd2 in schedule.commands())
        self.assertTrue(cmd3 in schedule.commands())
